@@ -9,23 +9,24 @@ from inspect import isclass
 from pickle import PicklingError
 from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Type, TypeVar, Union, cast
 
-import joblib
+from typing_extensions import Concatenate, ParamSpec
 
 from tpcp._base import _get_annotated_fields_of_type
+from tpcp._hash import custom_hash
 from tpcp._parameters import _ParaTypes
 from tpcp.exceptions import PotentialUserErrorWarning
 
 if TYPE_CHECKING:
     from tpcp import Algorithm, OptimizableAlgorithm, OptimizablePipeline
-    from tpcp._algorithm import Algorithm_
+    from tpcp._algorithm import AlgorithmT
 
-    Optimizable_ = TypeVar("Optimizable_", OptimizablePipeline, OptimizableAlgorithm)
+    OptimizableT = TypeVar("OptimizableT", OptimizablePipeline, OptimizableAlgorithm)
 
 ACTION_METHOD_INDICATOR = "__tpcp_action_method"
 OPTIMIZE_METHOD_INDICATOR = "__tpcp_optimize_method"
 
 
-R = TypeVar("R")
+P = ParamSpec("P")
 
 
 def get_action_method(instance: Algorithm, method_name: Optional[str] = None) -> Callable:
@@ -124,18 +125,18 @@ def is_action_applied(instance: Algorithm) -> bool:
     return True
 
 
-def _check_safe_run(algorithm: Algorithm_, old_method: Callable, *args: Any, **kwargs: Any) -> Algorithm_:
+def _check_safe_run(algorithm: AlgorithmT, old_method: Callable, *args: Any, **kwargs: Any) -> AlgorithmT:
     """Run the pipeline and check that run behaved as expected."""
     before_paras = algorithm.get_params()
-    before_paras_hash = joblib.hash(before_paras)
-    output: Algorithm_
+    before_paras_hash = custom_hash(before_paras)
+    output: AlgorithmT
     if hasattr(old_method, "__self__"):
         # In this case the method is already bound and we do not need to pass the algo as first argument
         output = old_method(*args, **kwargs)
     else:
         output = old_method(algorithm, *args, **kwargs)
     after_paras = algorithm.get_params()
-    after_paras_hash = joblib.hash(after_paras)
+    after_paras_hash = custom_hash(after_paras)
     if not before_paras_hash == after_paras_hash:
         raise ValueError(
             f"Running `{old_method.__name__}` of {type(algorithm).__name__} did modify the parameters of the "
@@ -163,7 +164,9 @@ def _check_safe_run(algorithm: Algorithm_, old_method: Callable, *args: Any, **k
     return output
 
 
-def make_action_safe(action_method: Callable[..., R]) -> Callable[..., R]:
+def make_action_safe(
+    action_method: Callable[Concatenate[AlgorithmT, P], AlgorithmT]
+) -> Callable[Concatenate[AlgorithmT, P], AlgorithmT]:
     """Mark a method as an "action" and apply a set of runtime checks to prevent implementation errors.
 
     This decorator marks a method as action.
@@ -198,7 +201,7 @@ def make_action_safe(action_method: Callable[..., R]) -> Callable[..., R]:
         return action_method
 
     @wraps(action_method)
-    def safe_wrapped(self: Algorithm_, *args: Any, **kwargs: Any) -> Algorithm_:
+    def safe_wrapped(self: AlgorithmT, *args: P.args, **kwargs: P.kwargs) -> AlgorithmT:
         if action_method.__name__ not in get_action_methods_names(self):
             warnings.warn(
                 "The `make_action_safe` decorator should only be applied to an action method "
@@ -236,7 +239,7 @@ def _get_nested_opti_paras(algorithm: Algorithm, opti_para_names: List[str]) ->
     return optimizable_paras, other_paras
 
 
-def _check_safe_optimize(algorithm: Optimizable_, old_method: Callable, *args: Any, **kwargs: Any) -> Optimizable_:
+def _check_safe_optimize(algorithm: OptimizableT, old_method: Callable, *args: Any, **kwargs: Any) -> OptimizableT:
     # record the hash of the pipeline to make an educated guess if the optimization works
     opti_para_names = _get_annotated_fields_of_type(algorithm, _ParaTypes.OPTI)
     optimizable_paras, other_paras = _get_nested_opti_paras(algorithm, opti_para_names)
@@ -246,9 +249,9 @@ def _check_safe_optimize(algorithm: Optimizable_, old_method: Callable, *args: A
             "Mark at least one; parameter with the `OptiPara`/`OptimizablePara` annotation to use "
             "`self_optimize`."
         )
-    before_hash_optimizable = joblib.hash(optimizable_paras)
-    before_hash_other = joblib.hash(other_paras)
-    optimized_algorithm: Optimizable_
+    before_hash_optimizable = custom_hash(optimizable_paras)
+    before_hash_other = custom_hash(other_paras)
+    optimized_algorithm: OptimizableT
     if hasattr(old_method, "__self__"):
         # In this case the method is already bound and we do not need to pass the algo as first argument
         optimized_algorithm = old_method(*args, **kwargs)
@@ -264,8 +267,8 @@ def _check_safe_optimize(algorithm: Optimizable_, old_method: Callable, *args: A
     # The first hash records any changes to the object.
     # The second hash only records changes to the parameters, because everything else is removed by clone.
     # Hence, if we see differences between the hashes, other things besides the parameters are changed.
-    after_hash = joblib.hash(optimized_algorithm)
-    after_hash_after_clone = joblib.hash(optimized_algorithm.clone())
+    after_hash = custom_hash(optimized_algorithm)
+    after_hash_after_clone = custom_hash(optimized_algorithm.clone())
     if after_hash_after_clone != after_hash:
         raise RuntimeError(
             "Optimizing seems to have changed class attributes that are not parameters (i.e. not provided in the "
@@ -280,15 +283,15 @@ def _check_safe_optimize(algorithm: Optimizable_, old_method: Callable, *args: A
     # It takes care of including and excluding the correct parameters in the other list, even if nested paras are
     # marked as "Optimizable"
     after_optimizable_paras, after_other_paras = _get_nested_opti_paras(algorithm, opti_para_names)
-    after_hash_optimizable = joblib.hash(after_optimizable_paras)
-    after_hash_other = joblib.hash(after_other_paras)
+    after_hash_optimizable = custom_hash(after_optimizable_paras)
+    after_hash_other = custom_hash(after_other_paras)
     if before_hash_other != after_hash_other:
         # In this case we raise an error anyway, so lets go deep:
         removed_paras = set(other_paras) - set(after_other_paras)
         added_paras = set(after_other_paras) - set(other_paras)
         changed_paras = []
         for k in set(other_paras) - set(removed_paras):
-            if joblib.hash(other_paras[k]) != joblib.hash(after_other_paras[k]):
+            if custom_hash(other_paras[k]) != custom_hash(after_other_paras[k]):
                 changed_paras.append(k)
         changed_paras = sorted(changed_paras)
         changed_paras.extend([f"{p} (removed)" for p in sorted(removed_paras)])
@@ -312,7 +315,9 @@ def _check_safe_optimize(algorithm: Optimizable_, old_method: Callable, *args: A
     return optimized_algorithm
 
 
-def make_optimize_safe(self_optimize_method: Callable[..., R]) -> Callable[..., R]:
+def make_optimize_safe(
+    self_optimize_method: Callable[Concatenate[OptimizableT, P], OptimizableT]
+) -> Callable[Concatenate[OptimizableT, P], OptimizableT]:
     """Apply a set of runtime checks to a custom `self_optimize` method to prevent implementation errors.
 
     The following things are checked:
@@ -351,7 +356,7 @@ def make_optimize_safe(self_optimize_method: Callable[..., R]) -> Callable[...,
         return self_optimize_method
 
     @wraps(self_optimize_method)
-    def safe_wrapped(self: Optimizable_, *args: Any, **kwargs: Any) -> Optimizable_:
+    def safe_wrapped(self: OptimizableT, *args: P.args, **kwargs: P.kwargs) -> OptimizableT:
         if self_optimize_method.__name__ != "self_optimize":
             warnings.warn(
                 "The `make_optimize_safe` decorator is only meant for the `self_optimize` method, but you applied it "
